/*-
 * HP-200LX Nikon-compatible Remote Control
 *
 * (C) Copyright 1992-2009 Diomidis Spinellis.  All rights reserved.
 *
 * Permission to use, copy, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: nikonrc.c,v 1.3 2009/08/12 20:27:34 dds Exp $
 *
 */

#include <stdio.h>
#include <time.h>

#include "serial.h"
#include "hpir.h"

int getopt(int argc, char *argv[], char *optstring);
extern char	*optarg;	/* Global argument pointer. */
extern int	optind;		/* Global argv index. */

static void
enable_serial_power(void)
{
	_asm {
		mov ax, 0x4a01
		int 0x15
		mov ax, 0x4901
		int 0x15
	}
}

static void
disable_serial_power(void)
{
	_asm {
		mov ax, 0x4900
		int 0x15
		mov ax, 0x4a00
		int 0x15
	}
}

static void
enable_light_sleep(void)
{
	_asm mov ax, 4e01h
	_asm int 15h
}

static void
disable_light_sleep(void)
{
	_asm mov ax, 4e00h
	_asm int 15h
}

static void
ir_start_send(unsigned int modfr)
{
	#define CLOCK 1840000
	unsigned int divisor = (unsigned int)((unsigned long)CLOCK / modfr);

	disable_light_sleep();
	enable_serial_power();
	_disable();
	outp(HIR, IRCNT);
	outp(HDR, 0);
	outp(HIR, IRFMAT);
	outp(HDR, MDLTE | MDSEL);
	outp(LCR, DLAB | WLS0 | WLS1);
	outp(DLM, divisor / 256);
	outp(DLL, divisor % 256);
	outp(LCR, 0);
}

static void
ir_start_receive(void)
{
	disable_light_sleep();
	enable_serial_power();
	_disable();
}

static void
ir_end(void)
{
	outp(HIR, IRCNT);
	outp(HDR, 0);
	outp(HIR, IRFMAT);
	outp(HDR, 0);
	outp(LCR, DLAB | WLS0 | WLS1);
	outp(DLM, 0);
	outp(DLL, 0);
	outp(LCR, 0);
	_enable();
	enable_light_sleep();
	disable_serial_power();
}

/*
 * Set the infrared red modulation frequency to modfr.
 * Send an array of n timings repeated the specified times.
 * Timing values are given in loop iterations and consist of an on time
 * followed by an off time.
 */
static void
play(unsigned int modfr, unsigned int *timings, int n, int repeat)
{
	int i, j, k, r;
	unsigned int *t;

	ir_start_send(modfr);
	outp(HIR, IRCNT);
	for (r = 0; r < repeat; r++) {
		t = timings;
		for (i = 0; i < n; i += 2) {
			/* High bit */
			k = *t++;
			outp(HDR, MDLD);
			for (j = 0; j < k; j++)
				_asm nop
			/* Low bit */
			outp(HDR, 0);
			k = *t++;
			for (j = 0; j < k; j++)
				_asm nop
		}
	}
	ir_end();
}

static void
usage()
{
	fprintf(stderr, "usage: nikonrc [-r repeat pause time] [-b bulb time]\n");
	exit(1);
}

/* Display msg and countdown for the specified number of seconds. */
static void
countdown(const char *msg, int nsecs)
{
	time_t start, end, oend;

	printf("%s: ", msg);
	time(&start);
	oend = 0;
	do {
		time(&end);
		if (oend != end) {
			printf("%ld ", nsecs - (end - start));
			oend = end;
		}
	} while (end - start < nsecs);
	putchar('\n');
}

/* Fire a single shot or for the specified bulb time. */
static void
fire(int bulb_time)
{
	/*
	 * Adjustment macro for HP 200LX timed with a stopwatch.
	 * (As interrupts are disabled the RTC doesn't work.)
	 * 1000 numbers of n=1000 timings take 10.19"s
	 * 1000 numbers of n=2000 timings take 19.48"s
	 * 1000 numbers of n=4000 timings take 38.60"s
	 * Therefore each timing lasts 9.65us
	 */
	#define A(x) ((int)(x / 9.65))
	/* See http://www.alanmacek.com/nikon/ */
	static unsigned int timings[] = {
		A(2000),	A(28000),
		 A(500),	 A(1500),
		 A(500),	 A(3500),
		 A(500), 	A(63000),
	};
	#define NVAL (sizeof(timings) / sizeof(unsigned int))

	if (bulb_time) {
		play(38000u, timings, NVAL, 2);
		countdown("Shooting", bulb_time);
		play(38000u, timings, NVAL, 2);
	} else
		play(38000u, timings, NVAL, 2);
}

int
main(int argc, char *argv[])
{
	int bulb_time = 0;
	int pause_time = 0;
	int c;

	while ((c = getopt(argc, argv, "b:r:")) != EOF)
		switch (c) {
		case 'b':
			if (!optarg)
				usage();
			bulb_time = atoi(optarg);
			break;
		case 'r':
			if (!optarg)
				usage();
			pause_time = atoi(optarg);
			break;
		case '?':
			usage();
		}

	if (pause_time)
		/* Repeat forever. */
		for (;;) {
			fire(bulb_time);
			countdown("Pausing", pause_time);
		}
	else
		fire(bulb_time);
	return 0;
}
